package org.antlr.analysis;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import org.antlr.analysis.DFAState;
import org.antlr.analysis.DecisionProbe;
import org.antlr.analysis.Label;
import org.antlr.analysis.NFA;
import org.antlr.analysis.NFAState;
import org.antlr.analysis.NFAToDFAConverter;
import org.antlr.analysis.NonLLStarDecisionException;
import org.antlr.analysis.SemanticContext;
import org.antlr.analysis.Transition;
import org.antlr.codegen.CodeGenerator;
import org.antlr.misc.IntSet;
import org.antlr.misc.IntervalSet;
import org.antlr.misc.Utils;
import org.antlr.runtime.IntStream;
import org.antlr.tool.ErrorManager;
import org.antlr.tool.FASerializer;
import org.antlr.tool.GrammarAST;
import org.antlr.tool.Interpreter;
import org.antlr.tool.Rule;
import org.stringtemplate.v4.ST;

public class DFA {
  public static final int REACHABLE_UNKNOWN = -2;
  
  public static final int REACHABLE_BUSY = -1;
  
  public static final int REACHABLE_NO = 0;
  
  public static final int REACHABLE_YES = 1;
  
  public static final int CYCLIC_UNKNOWN = -2;
  
  public static final int CYCLIC_BUSY = -1;
  
  public static final int CYCLIC_DONE = 0;
  
  public static int MAX_TIME_PER_DFA_CREATION = 1000;
  
  public static int MAX_STATE_TRANSITIONS_FOR_TABLE = 65534;
  
  public DFAState startState;
  
  public int decisionNumber = 0;
  
  public NFAState decisionNFAStartState;
  
  public String description;
  
  protected Map<DFAState, DFAState> uniqueStates = new HashMap<DFAState, DFAState>();
  
  protected Vector<DFAState> states = new Vector<DFAState>();
  
  protected int stateCounter = 0;
  
  protected int numberOfStates = 0;
  
  protected int user_k = -1;
  
  protected int max_k = -1;
  
  protected boolean reduced = true;
  
  protected boolean cyclic = false;
  
  public boolean predicateVisible = false;
  
  public boolean hasPredicateBlockedByAction = false;
  
  protected List<Integer> unreachableAlts;
  
  protected int nAlts = 0;
  
  protected DFAState[] altToAcceptState;
  
  public IntSet recursiveAltSet = new IntervalSet();
  
  public NFA nfa;
  
  protected NFAToDFAConverter nfaConverter;
  
  public DecisionProbe probe = new DecisionProbe(this);
  
  public Map<List<Integer>, Integer> edgeTransitionClassMap = new LinkedHashMap<List<Integer>, Integer>();
  
  protected int edgeTransitionClass = 0;
  
  public List<DFAState> specialStates;
  
  public List<ST> specialStateSTs;
  
  public Vector<Integer> accept;
  
  public Vector<Integer> eot;
  
  public Vector<Integer> eof;
  
  public Vector<Integer> min;
  
  public Vector<Integer> max;
  
  public Vector<Integer> special;
  
  public Vector<Vector<Integer>> transition;
  
  public Vector<Integer> transitionEdgeTables;
  
  protected int uniqueCompressedSpecialStateNum = 0;
  
  protected CodeGenerator generator = null;
  
  protected DFA() {}
  
  public DFA(int decisionNumber, NFAState decisionStartState) {
    this.decisionNumber = decisionNumber;
    this.decisionNFAStartState = decisionStartState;
    this.nfa = decisionStartState.nfa;
    this.nAlts = this.nfa.grammar.getNumberOfAltsForDecisionNFA(decisionStartState);
    initAltRelatedInfo();
    this.nfaConverter = new NFAToDFAConverter(this);
    try {
      this.nfaConverter.convert();
      verify();
      if (!this.probe.isDeterministic() || this.probe.analysisOverflowed())
        this.probe.issueWarnings(); 
      resetStateNumbersToBeContiguous();
    } catch (NonLLStarDecisionException nonLL) {
      this.probe.reportNonLLStarDecision(this);
      if (!okToRetryDFAWithK1())
        this.probe.issueWarnings(); 
    } 
  }
  
  public void resetStateNumbersToBeContiguous() {
    if (getUserMaxLookahead() > 0)
      return; 
    int snum = 0;
    for (int i = 0; i <= getMaxStateNumber(); i++) {
      DFAState s = getState(i);
      if (s != null) {
        boolean alreadyRenumbered = (s.stateNumber < i);
        if (!alreadyRenumbered) {
          s.stateNumber = snum;
          snum++;
        } 
      } 
    } 
    if (snum != getNumberOfStates())
      ErrorManager.internalError("DFA " + this.decisionNumber + ": " + this.decisionNFAStartState.getDescription() + " num unique states " + getNumberOfStates() + "!= num renumbered states " + snum); 
  }
  
  public List<? extends String> getJavaCompressedAccept() {
    return getRunLengthEncoding(this.accept);
  }
  
  public List<? extends String> getJavaCompressedEOT() {
    return getRunLengthEncoding(this.eot);
  }
  
  public List<? extends String> getJavaCompressedEOF() {
    return getRunLengthEncoding(this.eof);
  }
  
  public List<? extends String> getJavaCompressedMin() {
    return getRunLengthEncoding(this.min);
  }
  
  public List<? extends String> getJavaCompressedMax() {
    return getRunLengthEncoding(this.max);
  }
  
  public List<? extends String> getJavaCompressedSpecial() {
    return getRunLengthEncoding(this.special);
  }
  
  public List<List<? extends String>> getJavaCompressedTransition() {
    if (this.transition == null || this.transition.isEmpty())
      return null; 
    List<List<? extends String>> encoded = new ArrayList<List<? extends String>>(this.transition.size());
    for (int i = 0; i < this.transition.size(); i++) {
      Vector<Integer> transitionsForState = this.transition.elementAt(i);
      encoded.add(getRunLengthEncoding(transitionsForState));
    } 
    return encoded;
  }
  
  public List<? extends String> getRunLengthEncoding(List<Integer> data) {
    if (data == null || data.isEmpty()) {
      List<String> empty = new ArrayList<String>();
      empty.add("");
      return empty;
    } 
    int size = Math.max(2, data.size() / 2);
    List<String> encoded = new ArrayList<String>(size);
    int i = 0;
    Integer emptyValue = Utils.integer(-1);
    while (i < data.size()) {
      Integer I = data.get(i);
      if (I == null)
        I = emptyValue; 
      int n = 0;
      for (int j = i; j < data.size(); ) {
        Integer v = data.get(j);
        if (v == null)
          v = emptyValue; 
        if (I.equals(v)) {
          n++;
          j++;
        } 
      } 
      encoded.add(this.generator.target.encodeIntAsCharEscape((char)n));
      encoded.add(this.generator.target.encodeIntAsCharEscape((char)I.intValue()));
      i += n;
    } 
    return encoded;
  }
  
  public void createStateTables(CodeGenerator generator) {
    Iterator<DFAState> it;
    this.generator = generator;
    this.description = getNFADecisionStartState().getDescription();
    this.description = generator.target.getTargetStringLiteralFromString(this.description);
    this.special = new Vector<Integer>(getNumberOfStates());
    this.special.setSize(getNumberOfStates());
    this.specialStates = new ArrayList<DFAState>();
    this.specialStateSTs = new ArrayList<ST>();
    this.accept = new Vector<Integer>(getNumberOfStates());
    this.accept.setSize(getNumberOfStates());
    this.eot = new Vector<Integer>(getNumberOfStates());
    this.eot.setSize(getNumberOfStates());
    this.eof = new Vector<Integer>(getNumberOfStates());
    this.eof.setSize(getNumberOfStates());
    this.min = new Vector<Integer>(getNumberOfStates());
    this.min.setSize(getNumberOfStates());
    this.max = new Vector<Integer>(getNumberOfStates());
    this.max.setSize(getNumberOfStates());
    this.transition = new Vector<Vector<Integer>>(getNumberOfStates());
    this.transition.setSize(getNumberOfStates());
    this.transitionEdgeTables = new Vector<Integer>(getNumberOfStates());
    this.transitionEdgeTables.setSize(getNumberOfStates());
    if (getUserMaxLookahead() > 0) {
      it = this.states.iterator();
    } else {
      it = getUniqueStates().values().iterator();
    } 
    while (it.hasNext()) {
      DFAState s = it.next();
      if (s == null)
        continue; 
      if (s.isAcceptState()) {
        this.accept.set(s.stateNumber, Utils.integer(s.getUniquelyPredictedAlt()));
        continue;
      } 
      createMinMaxTables(s);
      createTransitionTableEntryForState(s);
      createSpecialTable(s);
      createEOTAndEOFTables(s);
    } 
    for (int i = 0; i < this.specialStates.size(); i++) {
      DFAState ss = this.specialStates.get(i);
      ST stateST = generator.generateSpecialState(ss);
      this.specialStateSTs.add(stateST);
    } 
  }
  
  protected void createMinMaxTables(DFAState s) {
    int smin = 65536;
    int smax = -3;
    for (int j = 0; j < s.getNumberOfTransitions(); j++) {
      Transition edge = s.transition(j);
      Label label = edge.label;
      if (label.isAtom()) {
        if (label.getAtom() >= 0) {
          if (label.getAtom() < smin)
            smin = label.getAtom(); 
          if (label.getAtom() > smax)
            smax = label.getAtom(); 
        } 
      } else if (label.isSet()) {
        IntervalSet labels = (IntervalSet)label.getSet();
        int lmin = labels.getMinElement();
        if (lmin < smin && lmin >= 0)
          smin = labels.getMinElement(); 
        if (labels.getMaxElement() > smax)
          smax = labels.getMaxElement(); 
      } 
    } 
    if (smax < 0) {
      smin = 0;
      smax = 0;
    } 
    this.min.set(s.stateNumber, Utils.integer((char)smin));
    this.max.set(s.stateNumber, Utils.integer((char)smax));
    if (smax < 0 || smin > 65535 || smin < 0)
      ErrorManager.internalError("messed up: min=" + this.min + ", max=" + this.max); 
  }
  
  protected void createTransitionTableEntryForState(DFAState s) {
    int smax = ((Integer)this.max.get(s.stateNumber)).intValue();
    int smin = ((Integer)this.min.get(s.stateNumber)).intValue();
    Vector<Integer> stateTransitions = new Vector<Integer>(smax - smin + 1);
    stateTransitions.setSize(smax - smin + 1);
    this.transition.set(s.stateNumber, stateTransitions);
    for (int j = 0; j < s.getNumberOfTransitions(); j++) {
      Transition edge = s.transition(j);
      Label label = edge.label;
      if (label.isAtom() && label.getAtom() >= 0) {
        int labelIndex = label.getAtom() - smin;
        stateTransitions.set(labelIndex, Utils.integer(edge.target.stateNumber));
      } else if (label.isSet()) {
        IntervalSet labels = (IntervalSet)label.getSet();
        int[] atoms = labels.toArray();
        for (int a = 0; a < atoms.length; a++) {
          if (atoms[a] >= 0) {
            int labelIndex = atoms[a] - smin;
            stateTransitions.set(labelIndex, Utils.integer(edge.target.stateNumber));
          } 
        } 
      } 
    } 
    Integer edgeClass = this.edgeTransitionClassMap.get(stateTransitions);
    if (edgeClass != null) {
      this.transitionEdgeTables.set(s.stateNumber, edgeClass);
    } else {
      edgeClass = Utils.integer(this.edgeTransitionClass);
      this.transitionEdgeTables.set(s.stateNumber, edgeClass);
      this.edgeTransitionClassMap.put(stateTransitions, edgeClass);
      this.edgeTransitionClass++;
    } 
  }
  
  protected void createEOTAndEOFTables(DFAState s) {
    for (int j = 0; j < s.getNumberOfTransitions(); j++) {
      Transition edge = s.transition(j);
      Label label = edge.label;
      if (label.isAtom()) {
        if (label.getAtom() == -2) {
          this.eot.set(s.stateNumber, Utils.integer(edge.target.stateNumber));
        } else if (label.getAtom() == -1) {
          this.eof.set(s.stateNumber, Utils.integer(edge.target.stateNumber));
        } 
      } else if (label.isSet()) {
        IntervalSet labels = (IntervalSet)label.getSet();
        int[] atoms = labels.toArray();
        for (int a = 0; a < atoms.length; a++) {
          if (atoms[a] == -2) {
            this.eot.set(s.stateNumber, Utils.integer(edge.target.stateNumber));
          } else if (atoms[a] == -1) {
            this.eof.set(s.stateNumber, Utils.integer(edge.target.stateNumber));
          } 
        } 
      } 
    } 
  }
  
  protected void createSpecialTable(DFAState s) {
    boolean hasSemPred = false;
    for (int j = 0; j < s.getNumberOfTransitions(); j++) {
      Transition edge = s.transition(j);
      Label label = edge.label;
      if (label.isSemanticPredicate() || ((DFAState)edge.target).getGatedPredicatesInNFAConfigurations() != null) {
        hasSemPred = true;
        break;
      } 
    } 
    int smax = ((Integer)this.max.get(s.stateNumber)).intValue();
    int smin = ((Integer)this.min.get(s.stateNumber)).intValue();
    if (hasSemPred || smax - smin > MAX_STATE_TRANSITIONS_FOR_TABLE) {
      this.special.set(s.stateNumber, Utils.integer(this.uniqueCompressedSpecialStateNum));
      this.uniqueCompressedSpecialStateNum++;
      this.specialStates.add(s);
    } else {
      this.special.set(s.stateNumber, Utils.integer(-1));
    } 
  }
  
  public int predict(IntStream input) {
    Interpreter interp = new Interpreter(this.nfa.grammar, input);
    return interp.predict(this);
  }
  
  protected DFAState addState(DFAState d) {
    if (getUserMaxLookahead() > 0)
      return d; 
    DFAState existing = this.uniqueStates.get(d);
    if (existing != null)
      return existing; 
    this.uniqueStates.put(d, d);
    this.numberOfStates++;
    return d;
  }
  
  public void removeState(DFAState d) {
    DFAState it = this.uniqueStates.remove(d);
    if (it != null)
      this.numberOfStates--; 
  }
  
  public Map<DFAState, DFAState> getUniqueStates() {
    return this.uniqueStates;
  }
  
  public int getMaxStateNumber() {
    return this.states.size() - 1;
  }
  
  public DFAState getState(int stateNumber) {
    return this.states.get(stateNumber);
  }
  
  public void setState(int stateNumber, DFAState d) {
    this.states.set(stateNumber, d);
  }
  
  public boolean isReduced() {
    return this.reduced;
  }
  
  public boolean isCyclic() {
    return (this.cyclic && getUserMaxLookahead() == 0);
  }
  
  public boolean isClassicDFA() {
    return (!isCyclic() && !this.nfa.grammar.decisionsWhoseDFAsUsesSemPreds.contains(this) && !this.nfa.grammar.decisionsWhoseDFAsUsesSynPreds.contains(this));
  }
  
  public boolean canInlineDecision() {
    return (!isCyclic() && !this.probe.isNonLLStarDecision() && getNumberOfStates() < CodeGenerator.MAX_ACYCLIC_DFA_STATES_INLINE);
  }
  
  public boolean isTokensRuleDecision() {
    if (this.nfa.grammar.type != 1)
      return false; 
    NFAState nfaStart = getNFADecisionStartState();
    Rule r = this.nfa.grammar.getLocallyDefinedRule("Tokens");
    NFAState TokensRuleStart = r.startState;
    NFAState TokensDecisionStart = (NFAState)(TokensRuleStart.transition[0]).target;
    return (nfaStart == TokensDecisionStart);
  }
  
  public int getUserMaxLookahead() {
    if (this.user_k >= 0)
      return this.user_k; 
    this.user_k = this.nfa.grammar.getUserMaxLookahead(this.decisionNumber);
    return this.user_k;
  }
  
  public boolean getAutoBacktrackMode() {
    return this.nfa.grammar.getAutoBacktrackMode(this.decisionNumber);
  }
  
  public void setUserMaxLookahead(int k) {
    this.user_k = k;
  }
  
  public int getMaxLookaheadDepth() {
    if (hasCycle())
      return Integer.MAX_VALUE; 
    return _getMaxLookaheadDepth(this.startState, 0);
  }
  
  int _getMaxLookaheadDepth(DFAState d, int depth) {
    int max = depth;
    for (int i = 0; i < d.getNumberOfTransitions(); i++) {
      Transition t = d.transition(i);
      if (!t.isSemanticPredicate()) {
        DFAState edgeTarget = (DFAState)t.target;
        int m = _getMaxLookaheadDepth(edgeTarget, depth + 1);
        max = Math.max(max, m);
      } 
    } 
    return max;
  }
  
  public boolean hasSynPred() {
    boolean has = _hasSynPred(this.startState, new HashSet<DFAState>());
    return has;
  }
  
  public boolean getHasSynPred() {
    return hasSynPred();
  }
  
  boolean _hasSynPred(DFAState d, Set<DFAState> busy) {
    busy.add(d);
    for (int i = 0; i < d.getNumberOfTransitions(); i++) {
      Transition t = d.transition(i);
      if (t.isSemanticPredicate()) {
        SemanticContext ctx = t.label.getSemanticContext();
        if (ctx.isSyntacticPredicate())
          return true; 
      } 
      DFAState edgeTarget = (DFAState)t.target;
      if (!busy.contains(edgeTarget) && _hasSynPred(edgeTarget, busy))
        return true; 
    } 
    return false;
  }
  
  public boolean hasSemPred() {
    boolean has = _hasSemPred(this.startState, new HashSet<DFAState>());
    return has;
  }
  
  boolean _hasSemPred(DFAState d, Set<DFAState> busy) {
    busy.add(d);
    for (int i = 0; i < d.getNumberOfTransitions(); i++) {
      Transition t = d.transition(i);
      if (t.isSemanticPredicate()) {
        SemanticContext ctx = t.label.getSemanticContext();
        if (ctx.hasUserSemanticPredicate())
          return true; 
      } 
      DFAState edgeTarget = (DFAState)t.target;
      if (!busy.contains(edgeTarget) && _hasSemPred(edgeTarget, busy))
        return true; 
    } 
    return false;
  }
  
  public boolean hasCycle() {
    boolean cyclic = _hasCycle(this.startState, new HashMap<DFAState, Integer>());
    return cyclic;
  }
  
  boolean _hasCycle(DFAState d, Map<DFAState, Integer> busy) {
    busy.put(d, Integer.valueOf(-1));
    for (int i = 0; i < d.getNumberOfTransitions(); i++) {
      Transition t = d.transition(i);
      DFAState target = (DFAState)t.target;
      int cond = -2;
      if (busy.get(target) != null)
        cond = ((Integer)busy.get(target)).intValue(); 
      if (cond == -1)
        return true; 
      if (cond != 0 && _hasCycle(target, busy))
        return true; 
    } 
    busy.put(d, Integer.valueOf(0));
    return false;
  }
  
  public List<Integer> getUnreachableAlts() {
    return this.unreachableAlts;
  }
  
  public void verify() {
    doesStateReachAcceptState(this.startState);
  }
  
  protected boolean doesStateReachAcceptState(DFAState d) {
    if (d.isAcceptState()) {
      d.setAcceptStateReachable(1);
      int predicts = d.getUniquelyPredictedAlt();
      this.unreachableAlts.remove(Utils.integer(predicts));
      return true;
    } 
    d.setAcceptStateReachable(-1);
    boolean anEdgeReachesAcceptState = false;
    for (int i = 0; i < d.getNumberOfTransitions(); i++) {
      Transition t = d.transition(i);
      DFAState edgeTarget = (DFAState)t.target;
      int targetStatus = edgeTarget.getAcceptStateReachable();
      if (targetStatus == -1) {
        this.cyclic = true;
      } else if (targetStatus == 1) {
        anEdgeReachesAcceptState = true;
      } else if (targetStatus != 0) {
        if (doesStateReachAcceptState(edgeTarget))
          anEdgeReachesAcceptState = true; 
      } 
    } 
    if (anEdgeReachesAcceptState) {
      d.setAcceptStateReachable(1);
    } else {
      d.setAcceptStateReachable(0);
      this.reduced = false;
    } 
    return anEdgeReachesAcceptState;
  }
  
  public void findAllGatedSynPredsUsedInDFAAcceptStates() {
    int nAlts = getNumberOfAlts();
    for (int i = 1; i <= nAlts; i++) {
      DFAState a = getAcceptState(i);
      if (a != null) {
        Set<? extends SemanticContext> synpreds = a.getGatedSyntacticPredicatesInNFAConfigurations();
        if (synpreds != null)
          for (SemanticContext semctx : synpreds)
            this.nfa.grammar.synPredUsedInDFA(this, semctx);  
      } 
    } 
  }
  
  public NFAState getNFADecisionStartState() {
    return this.decisionNFAStartState;
  }
  
  public DFAState getAcceptState(int alt) {
    return this.altToAcceptState[alt];
  }
  
  public void setAcceptState(int alt, DFAState acceptState) {
    this.altToAcceptState[alt] = acceptState;
  }
  
  public String getDescription() {
    return this.description;
  }
  
  public int getDecisionNumber() {
    return this.decisionNFAStartState.getDecisionNumber();
  }
  
  public boolean okToRetryDFAWithK1() {
    boolean nonLLStarOrOverflowAndPredicateVisible = ((this.probe.isNonLLStarDecision() || this.probe.analysisOverflowed()) && this.predicateVisible);
    return (getUserMaxLookahead() != 1 && nonLLStarOrOverflowAndPredicateVisible);
  }
  
  public String getReasonForFailure() {
    StringBuilder buf = new StringBuilder();
    if (this.probe.isNonLLStarDecision()) {
      buf.append("non-LL(*)");
      if (this.predicateVisible)
        buf.append(" && predicate visible"); 
    } 
    if (this.probe.analysisOverflowed()) {
      buf.append("recursion overflow");
      if (this.predicateVisible)
        buf.append(" && predicate visible"); 
    } 
    buf.append("\n");
    return buf.toString();
  }
  
  public GrammarAST getDecisionASTNode() {
    return this.decisionNFAStartState.associatedASTNode;
  }
  
  public boolean isGreedy() {
    GrammarAST blockAST = this.nfa.grammar.getDecisionBlockAST(this.decisionNumber);
    Object v = this.nfa.grammar.getBlockOption(blockAST, "greedy");
    if (v != null && v.equals("false"))
      return false; 
    return true;
  }
  
  public DFAState newState() {
    DFAState n = new DFAState(this);
    n.stateNumber = this.stateCounter;
    this.stateCounter++;
    this.states.setSize(n.stateNumber + 1);
    this.states.set(n.stateNumber, n);
    return n;
  }
  
  public int getNumberOfStates() {
    if (getUserMaxLookahead() > 0)
      return this.states.size(); 
    return this.numberOfStates;
  }
  
  public int getNumberOfAlts() {
    return this.nAlts;
  }
  
  protected void initAltRelatedInfo() {
    this.unreachableAlts = new LinkedList<Integer>();
    for (int i = 1; i <= this.nAlts; i++)
      this.unreachableAlts.add(Utils.integer(i)); 
    this.altToAcceptState = new DFAState[this.nAlts + 1];
  }
  
  public String toString() {
    FASerializer serializer = new FASerializer(this.nfa.grammar);
    if (this.startState == null)
      return ""; 
    return serializer.serialize(this.startState, false);
  }
}
